@@ -12,7 +12,7 @@
 
 """SSH Module for connecting to and automating remote commands.
 
-Supports proxying, as in `ssh -A`
+Supports proxying through an ssh tunnel ('gateway' keyword argument.)
 
 To control the behavior of the SSH client, use the specific connect_with_*
 calls. The .connect() call behaves like the ssh command and attempts a number
@@ -31,13 +31,13 @@ import getpass
 import logging
 import os
 import re
-import tempfile
 import time
 
 import paramiko
 import six
 
 from satori import errors
+from satori import utils
 
 LOG = logging.getLogger(__name__)
 MIN_PASSWORD_PROMPT_LEN = 8
@@ -47,6 +47,7 @@ TTY_REQUIRED = [
     "you must have a tty to run sudo",
     "is not a tty",
     "no tty present",
+    "must be run from a terminal",
 ]
 
 
@@ -99,9 +100,9 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
     """Connects to devices via SSH to execute commands."""
 
     # pylint: disable=R0913
-    def __init__(self, host, password=None, username="root",
-                 private_key=None, key_filename=None, port=22,
-                 timeout=20, proxy=None, options=None, interactive=False):
+    def __init__(self, host, password=None, username="root", private_key=None,
+                 root_password=None, key_filename=None, port=22,
+                 timeout=20, gateway=None, options=None, interactive=False):
         """Create an instance of the SSH class.
 
         :param str host:        The ip address or host name of the server
@@ -109,13 +110,17 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
         :param str password:    A password to use for authentication
                                 or for unlocking a private key
         :param username:        The username to authenticate as
+        :param root_password:   root user password to be used if 'username'
+                                is not root. This will use 'username' and
+                                'password to login and then 'su' to root
+                                using root_password
         :param private_key:     Private SSH Key string to use
                                 (instead of using a filename)
         :param key_filename:    a private key filename (path)
         :param port:            tcp/ip port to use (defaults to 22)
         :param float timeout:   an optional timeout (in seconds) for the
                                 TCP connection
-        :param socket proxy:    an existing SSH instance to use
+        :param socket gateway:    an existing SSH instance to use
                                 for proxying
         :param dict options:    A dictionary used to set ssh options
                                 (when proxying).
@@ -129,25 +134,34 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
         """
         self.password = password
         self.host = host
-        self.username = username
+        self.username = username or 'root'
+        self.root_password = root_password
         self.private_key = private_key
         self.key_filename = key_filename
-        self.port = port
+        self.port = port or 22
         self.timeout = timeout
         self._platform_info = None
         self.options = options or {}
-        self.proxy = proxy
+        self.gateway = gateway
         self.sock = None
         self.interactive = interactive
 
-        if self.proxy:
-            if not isinstance(self.proxy, SSH):
-                raise TypeError("'proxy' must be a satori.ssh.SSH instance. "
+        self.escalation_command = 'sudo -i %s'
+        if self.root_password:
+            self.escalation_command = "su -c '%s'"
+
+        if self.gateway:
+            if not isinstance(self.gateway, SSH):
+                raise TypeError("'gateway' must be a satori.ssh.SSH instance. "
                                 "( instances of this type are returned by "
                                 "satori.ssh.connect() )")
 
         super(SSH, self).__init__()
 
+    def __del__(self):
+        """Destructor to close the connection."""
+        self.close()
+
     @classmethod
     def get_client(cls, *args, **kwargs):
         """Return an ssh client object from this module."""
@@ -155,20 +169,31 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
 
     @property
     def platform_info(self):
-        """Return distro, version, architecture."""
-        if not self._platform_info:
-            command = ('python -c '
-                       '"""import sys,platform as p;'
-                       'plat=list(p.dist()+(p.machine(),));'
-                       'sys.stdout.write(str(plat))"""')
+        """Return distro, version, architecture.
 
+        Requires >= Python 2.4 on remote system.
+        """
+        if not self._platform_info:
+            platform_command = "import platform,sys\n"
+            platform_command += utils.get_source_definition(
+                utils.get_platform_info)
+            platform_command += ("\nsys.stdout.write(str("
+                                 "get_platform_info()))\n")
+            command = 'echo -e """%s""" | python' % platform_command
             output = self.remote_execute(command)
             stdout = re.split('\n|\r\n', output['stdout'])[-1].strip()
-            plat = ast.literal_eval(stdout)
-            self._platform_info = {'dist': plat[0].lower(), 'version': plat[1],
-                                   'arch': plat[3]}
-
-        LOG.debug("Remote platform info: %s", self._platform_info)
+            if stdout:
+                try:
+                    plat = ast.literal_eval(stdout)
+                except SyntaxError as exc:
+                    plat = {'dist': 'unknown'}
+                    LOG.warning("Error parsing response from host '%s': %s",
+                                self.host, output, exc_info=exc)
+            else:
+                plat = {'dist': 'unknown'}
+                LOG.warning("Blank response from host '%s': %s",
+                            self.host, output)
+            self._platform_info = plat
         return self._platform_info
 
     def connect_with_host_keys(self):
@@ -219,13 +244,16 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
         """Set up client and connect to target."""
         self.load_system_host_keys()
 
-        if self.proxy:
-            # lazy load
-            self.sock = self._get_proxy_socket(self.proxy)
-
         if self.options.get('StrictHostKeyChecking') in (False, "no"):
             self.set_missing_host_key_policy(AcceptMissingHostKey())
 
+        if self.gateway:
+            # lazy load
+            if not self.gateway.get_transport():
+                self.gateway.connect()
+            self.sock = self.gateway.get_transport().open_channel(
+                'direct-tcpip', (self.host, self.port), ('', 0))
+
         return super(SSH, self).connect(
             self.host,
             timeout=kwargs.pop('timeout', self.timeout),
@@ -245,6 +273,11 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
         - username/password (will prompt if the password is not supplied and
                              interactive is true)
         """
+        # idempotency
+        if self.get_transport():
+            if self.get_transport().is_active():
+                return
+
         if self.private_key:
             try:
                 return self.connect_with_key()
@@ -300,10 +333,19 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
         finally:
             self.close()
 
+    def close(self):
+        """Close the connection to the remote host.
+
+        If an ssh tunnel is being used, close that first.
+        """
+        if self.gateway:
+            self.gateway.close()
+        return super(SSH, self).close()
+
     def _handle_tty_required(self, results, get_pty):
         """Determine whether the result implies a tty request."""
         if any(m in str(k) for m in TTY_REQUIRED for k in results.values()):
-            LOG.info('%s requires TTY for sudo. Using TTY mode.',
+            LOG.info('%s requires TTY for sudo/su. Using TTY mode.',
                      self.host)
             if get_pty is True:  # if this is *already* True
                 raise errors.GetPTYRetryFailure(
@@ -313,7 +355,7 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
                 return True
         return False
 
-    def _handle_password_prompt(self, stdin, stdout):
+    def _handle_password_prompt(self, stdin, stdout, su_auth=False):
         """Determine whether the remote host is prompting for a password.
 
         Respond to the prompt through stdin if applicable.
@@ -329,7 +371,11 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
                     LOG.warning("%s@%s encountered prompt! of length "
                                 " [%s] {%s}",
                                 self.username, self.host, buflen, prompt)
-                    stdin.write("%s\n" % self.password)
+                    if su_auth:
+                        LOG.warning("Escalating using 'su -'.")
+                        stdin.write("%s\n" % self.root_password)
+                    else:
+                        stdin.write("%s\n" % self.password)
                     stdin.flush()
                     return True
                 else:
@@ -340,8 +386,21 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
 
         return False
 
-    def remote_execute(self, command, with_exit_code=False,
-                       get_pty=False, wd=None):
+    def _command_is_already_running(self, command):
+        """Check to see if the command is already running using ps & grep."""
+        # check plain 'command' w/o prefix or escalation
+        check_cmd = 'ps -ef |grep -v grep|grep -c "%s"' % command
+        result = self.remote_execute(check_cmd, keepalive=True,
+                                     allow_many=True)
+        if result['stdout'] != '0':
+            return True
+        else:
+            LOG.debug("Remote command %s IS NOT already running. "
+                      "Continuing with remote_execute.", command)
+
+    def remote_execute(self, command, with_exit_code=False,  # noqa
+                       get_pty=False, cwd=None, keepalive=True,
+                       escalate=False, allow_many=True, **kw):
         """Execute an ssh command on a remote host.
 
         Tries cert auth first and falls back
@@ -349,123 +408,89 @@ class SSH(paramiko.SSHClient):  # pylint: disable=R0902
 
         :param command:         Shell command to be executed by this function.
         :param with_exit_code:  Include the exit_code in the return body.
-        :param wd:              The child's current directory will be changed
-                                to `wd` before it is executed. Note that this
+        :param cwd:             The child's current directory will be changed
+                                to `cwd` before it is executed. Note that this
                                 directory is not considered when searching the
                                 executable, so you can't specify the program's
                                 path relative to this argument
         :param get_pty:         Request a pseudo-terminal from the server.
+        :param allow_many:      If False, do not run command if it is already
+                                found running on remote client.
 
         :returns: a dict with stdin, stdout,
                   and (optionally) the exit code of the call.
         """
-        if wd:
-            prefix = "cd %s && " % wd
-            command = prefix + command
+        if escalate and self.username != 'root':
+            run_command = self.escalation_command % command
+        else:
+            run_command = command
 
-        LOG.debug("Executing '%s' on ssh://%s@%s:%s.",
-                  command, self.username, self.host, self.port)
+        if cwd:
+            prefix = "cd %s && " % cwd
+            run_command = prefix + run_command
+
+        # _command_is_already_running wont be called if allow_many is True
+        # python is great :)
+        if not allow_many and self._command_is_already_running(command):
+            raise errors.SatoriDuplicateCommandException(
+                "Remote command %s is already running and allow_many was "
+                "set to False. Aborting remote_execute." % command)
         try:
             self.connect()
-
             results = None
             chan = self.get_transport().open_session()
+            su_auth = False
+            if 'su -' in run_command:
+                su_auth = True
+                get_pty = True
             if get_pty:
                 chan.get_pty()
             stdin = chan.makefile('wb')
             stdout = chan.makefile('rb')
             stderr = chan.makefile_stderr('rb')
-            chan.exec_command(command)
+            LOG.debug("Executing '%s' on ssh://%s@%s:%s.",
+                      run_command, self.username, self.host, self.port)
+            chan.exec_command(run_command)
             LOG.debug('ssh://%s@%s:%d responded.', self.username, self.host,
                       self.port)
 
             time.sleep(.25)
-            self._handle_password_prompt(stdin, stdout)
+            self._handle_password_prompt(stdin, stdout, su_auth=su_auth)
 
             results = {
                 'stdout': stdout.read().strip(),
                 'stderr': stderr.read()
             }
 
-            LOG.debug("STDOUT from ssh://%s@%s:%d: %s",
+            LOG.debug("STDOUT from ssh://%s@%s:%d: %.5000s ...",
                       self.username, self.host, self.port,
                       results['stdout'])
-            LOG.debug("STDERR from ssh://%s@%s:%d: %s",
+            LOG.debug("STDERR from ssh://%s@%s:%d: %.5000s ...",
                       self.username, self.host, self.port,
                       results['stderr'])
             exit_code = chan.recv_exit_status()
 
             if with_exit_code:
                 results.update({'exit_code': exit_code})
-            chan.close()
+            if not keepalive:
+                chan.close()
 
             if self._handle_tty_required(results, get_pty):
                 return self.remote_execute(
-                    command, with_exit_code=with_exit_code, get_pty=True)
+                    command, with_exit_code=with_exit_code, get_pty=True,
+                    cwd=cwd, keepalive=keepalive, escalate=escalate,
+                    allow_many=allow_many)
 
             return results
 
         except Exception as exc:
-            LOG.info("ssh://%s@%s:%d failed.  %s", self.username, self.host,
+            LOG.info("ssh://%s@%s:%d failed. | %s", self.username, self.host,
                      self.port, exc)
             raise
         finally:
-            self.close()
+            if not keepalive:
+                self.close()
 
-    def _get_proxy_socket(self, proxy):
-        """Return a wrapped subprocess running ProxyCommand-driven programs.
-
-        Create a new CommandProxy instance.
-        Can be created from an existing SSH instance.
-        For proxy clients, please specify a private key filename.
-
-        To use an ssh proxy, you must use an SSH Key,
-        since a ProxyCommand cannot be passed a password.
-        """
-        if proxy.password:
-            LOG.warning("Proxying through a client which is authorized by "
-                        "a password is not currently implemented. Please "
-                        "use an ssh key.")
-
-        proxy.load_system_host_keys()
-        if proxy.options.get('StrictHostKeyChecking') in (False, "no"):
-            proxy.set_missing_host_key_policy(AcceptMissingHostKey())
-
-        if proxy.private_key and not proxy.key_filename:
-            tempkeyfile = tempfile.NamedTemporaryFile(
-                mode='w+', prefix=TEMPFILE_PREFIX,
-                dir=os.path.expanduser('~/'), delete=True)
-            tempkeyfile.write(proxy.private_key)
-            proxy.key_filename = tempkeyfile.name
-
-        pxd = {
-            'bastion': proxy.host,
-            'user': proxy.username,
-            'port': '-p %s' % proxy.port,
-            'options': ('-o ConnectTimeout=%s ' % proxy.timeout),
-            'target_host': self.host,
-            'target_port': self.port,
-        }
-        proxycommand = "ssh {options} -A {user}@{bastion} "
-
-        if proxy.key_filename:
-            proxy.key_filename = os.path.expanduser(proxy.key_filename)
-            proxy.key_filename = os.path.abspath(proxy.key_filename)
-            pxd.update({'identity': '-i %s' % proxy.key_filename})
-            proxycommand += "{identity} "
-
-        if proxy.options:
-            for key, val in sorted(proxy.options.items()):
-                if isinstance(val, bool):
-                    # turns booleans into `ssh -o` compat "yes" or "no"
-                    if val is True:
-                        val = "yes"
-                    if val is False:
-                        val = "no"
-                pxd['options'] += '-o %s=%s ' % (key, val)
-
-        proxycommand += "nc {target_host} {target_port}"
-        return paramiko.ProxyCommand(proxycommand.format(**pxd))
 
 # Share SSH.__init__'s docstring
 connect.__doc__ = SSH.__init__.__doc__
